%% tclp__arity.pl 
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                              TCLP                                         %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__arity.pl,v 1.1 2002/10/17 15:30:17 coquery Exp $

%% Arity predicates. Based on a chr solver with constraints on the arity of
%% type constructors.

:- module(tclp__arity, [unifyToArity/1,
                        find_mapping/3,
                        arity/2,
                        match/3,
                        match/5,
                        fold/5,
                        fold/7]).

:- use_module(tclp(tclp__utils)).
:- use_module(library(chr)).
:- use_module(library(lists)).

%% :- type arity.
%% :- pred arity(arity,atom,int).
%% :- pred arity(atom,int).

%% The solver, a simple one

handler arity.

constraints arity/3,%% arity(A,Cons,N) says that the N-th argument of
                    %% type constructor Cons as A as its arity
            arity/2 %% arity(Cons,N) says that the number of arguments
                    %% of Cons is N
.

%% Rules

%% Unifies the arity according to the constructor and the argument
consult1 @ arity(A,Cons,N) \ arity(B,Cons,N)
       <=> nonvar(Cons), nonvar(N) | A=B.

%% Unifies the number of arguments
consult2 @ arity(Cons,N1) \ arity(Cons,N2) <=> N1=N2.

%% Unifies the arg number according to the constructor and the arity
%% consult2 @ arity(A,Cons,N1) \ arity(A,Cons,N2)
%%       <=> nonvar(Cons), (nonvar(N1),var(N2);nonvar(N2),var(N1)) | N1=N2.


%% Predicates

%% Unifies the argument of the given type constructor with their arity
%% :- pred unifyToArity(type_term).
unifyToArity(Type) :-
        Type =.. [ TCons | Args ],
        unifyArgsToArity(Args,TCons,1).

unifyArgsToArity([],_,_).
unifyArgsToArity([A|As], TCons, N) :-
        arity(A, TCons, N),
        N1 is N+1,
        unifyArgsToArity(As, TCons, N1).


%% Mapping is a list of pair of integers such that for each pair M-N,
%% the M-th argument of Type1 has the same arity than the N-th argument of
%% Type2. find_mapping is memoized to avoid recomputation
%% :- pred find_mapping(type_term, type_term, list(pair(int,int))).
:- dynamic find_mapping/3.
find_mapping(Type1,Type2,Mapping) :-
        functor(Type1,TCons1, Nargs1),
        functor(Type2,TCons2, Nargs2),
        functor(Type1b,TCons1, Nargs1),
        functor(Type2b,TCons2, Nargs2),
        unifyToArity(Type1b),
        unifyToArity(Type2b),
        Type1b =.. [ TCons1 | ArgsArity1 ],
        Type2b =.. [ TCons2 | ArgsArity2 ],
        find_mapping_2(ArgsArity1, 1, ArgsArity2, [], Mapping),!,
        functor(Type1c,TCons1, Nargs1),
        functor(Type2c,TCons2, Nargs2),
        asserta((find_mapping(Type1c, Type2c, Mapping) :- !)).
        

find_mapping_2([],_,_,Mapping, Mapping).
find_mapping_2([X|Xs], NX, Ys, CurrentMapping, FinalMapping) :-
        find_mapping_3(X, NX, Ys, 1, CurrentMapping, NewMapping),
        NX1 is NX+1,
        find_mapping_2(Xs, NX1, Ys, NewMapping, FinalMapping).

find_mapping_3(_, _, [], _, Mapping, Mapping).
find_mapping_3(X, NX, [Y|Ys], NY, CurrentMapping, FinalMapping) :-
        (
          X == Y , NewMapping = [ NX-NY | CurrentMapping ]
        ;
          NewMapping = CurrentMapping
        ),
        NY1 is NY+1,
        find_mapping_3(X,NX,Ys,NY1, NewMapping, FinalMapping).

%% given two nonvar types, applies the given predicate to args with the same
%% arity in the types
%% Ex:
%% match(list(A),list(B),p) applied p(A,B)
%% :- pred match(type_term, type_term, atom).
match(Type1, Type2, Pred) :-
        find_mapping(Type1, Type2, Mapping),
        Type1 =.. [ _ | Args1 ],
        Type2 =.. [ _ | Args2 ],
        iter_match(Args1, Args2, Mapping, Pred).

iter_match(_,_,[],_).
iter_match(Args1, Args2, [ N1-N2 | MappingTail ], Pred) :-
        nth(N1, Args1, A1), nth(N2, Args2, A2),
        make_goal(Pred, [ A1, A2 ], Goal), call(Goal),
        iter_match(Args1, Args2, MappingTail, Pred).

%% given three nonvar types, applies one of the two given predicates to some
%% of their arguments, which arity is the same.
%% match(Type1, Type2, Type3, Pred2, Pred3)
%% for each argument A of Type3,
%% if there is an argument A1 in Type1 which has the same arity
%%   if there is an argument A2 in Type2 which has the same arity
%%      call Pred3(A1,A2,A)
%%   else
%%      call Pred2(A1,A)
%% else
%%   if there is an argument A2 in Type2 which has the same arity
%%      call Pred2(A2,A)
%%   else fail
%% :- pred match(type_term, type_term, type_term, atom, atom).
match(Type1, Type2, Type3, Pred2, Pred3) :-
        Type1 =.. [ _ | Args1 ],
        Type2 =.. [ _ | Args2 ],
        Type3 =.. [ _ | Args3 ],
        find_mapping(Type1, Type3, Mapping13),
        find_mapping(Type2, Type3, Mapping23),
        iter_match(Args1, Args2, Args3, 1, Mapping13, Mapping23, Pred2, Pred3).

iter_match(_,_,[],_,_,_,_,_).
iter_match(Args1, Args2, [A3| Args3], Nth,
           Mapping13, Mapping23, Pred2, Pred3) :-
        (
          member(AN1-Nth,Mapping13)
        ->
          nth(AN1, Args1, A1),
          (
            member(AN2-Nth, Mapping23)
          ->
            nth(AN2, Args2, A2),
            make_goal(Pred3, [ A1, A2, A3 ], Goal)
            %Goal =.. [ Pred3, A1, A2, A3 ]
          ;
            make_goal(Pred2, [ A1, A3 ], Goal)
            %Goal =.. [ Pred2, A1, A3 ]
          )
        ;
          (
            member(AN2-Nth, Mapping23)
          ->
            nth(AN2, Args2, A2),
            make_goal(Pred2, [ A2, A3 ], Goal)
            %Goal =.. [ Pred2, A2, A3 ]
          ;
            Goal = fail
          )
        ),
        call(Goal),
        Nth1 is Nth+1,
        iter_match(Args1, Args2, Args3, Nth1,
                   Mapping13, Mapping23, Pred2, Pred3).


%% same as match/3 except we have an accumulator and a result.
%% Ex:
%% fold(heap(A,B), heap(B,C), Pred, Accu, Result)
%% is equivalent to
%% Pred(A,C,Accu,Accu1), Pred(B,D,Accu1,Accu2), Accu2 = Result.
%%
%% :- pred fold(type_term, type_term, atom, A, A).

fold(Type1, Type2, Pred, Accu, Result) :-
        find_mapping(Type1, Type2, Mapping),
        Type1 =.. [ _ | Args1 ],
        Type2 =.. [ _ | Args2 ],
        iter_fold(Args1, Args2, Mapping, Pred, Accu, Result).

iter_fold(_,_, [], _, Accu, Accu).
iter_fold(Args1, Args2, [ N1 - N2 | MappingTail ], Pred, Accu, Result) :-
        nth(N1, Args1, A1), nth(N2, Args2, A2),
        make_goal(Pred, [ A1, A2, Accu, Accu1 ], Goal),
        %Goal =.. [ Pred, A1, A2, Accu, Accu1 ],
        call(Goal),
        iter_fold(Args1, Args2, MappingTail, Pred, Accu1, Result).

%% fold/7 is to match/5 what fold/5 is to match/3
%% That is, we have an accumulation of some result
%%
%% :- pred fold(type_term, type_term, type_term, atom, atom, A, A).

fold(Type1, Type2, Type3, Pred2, Pred3, Accu, Result) :-
        Type1 =.. [ _ | Args1 ],
        Type2 =.. [ _ | Args2 ],
        Type3 =.. [ _ | Args3 ],
        find_mapping(Type1, Type3, Mapping13),
        find_mapping(Type2, Type3, Mapping23),
        iter_fold(Args1, Args2, Args3, 1,
                  Mapping13, Mapping23, Pred2, Pred3,
                  Accu, Result).

iter_fold(_,_,[],_,
          _,_,_,_,
          Accu, Accu).
iter_fold(Args1, Args2, [ A3 | Args3 ], Nth,
          Mapping13, Mapping23, Pred2, Pred3,
          Accu, Result) :-
        (
          member(AN1-Nth,Mapping13)
        ->
          nth(AN1, Args1, A1),
          (
            member(AN2-Nth, Mapping23)
          ->
            nth(AN2, Args2, A2),
            make_goal(Pred3, [ A1, A2, A3, Accu, Accu1 ], Goal)
            %Goal =.. [ Pred3, A1, A2, A3, Accu, Accu1 ]
          ;
            make_goal(Pred2, [ A1, A3, Accu, Accu1 ], Goal)
            %Goal =.. [ Pred2, A1, A3, Accu, Accu1 ]
          )
        ;
          (
            member(AN2-Nth, Mapping23)
          ->
            nth(AN2, Args2, A2),
            make_goal(Pred2, [ A2, A3, Accu, Accu1 ], Goal)
            %Goal =.. [ Pred2, A2, A3, Accu, Accu1 ]
          ;
            Goal = fail
          )
        ),
        call(Goal),
        Nth1 is Nth+1,
        iter_fold(Args1, Args2, Args3, Nth1,
                  Mapping13, Mapping23, Pred2, Pred3,
                  Accu1, Result).

%% make_goal(Pred, Args, Goal)
%% Constructs a goal Goal using Pred as a predicate and args as its arguments
%% :- pred make_goal(term, list(term), goal).
make_goal(M:Pred, Args, M:Goal) :- !,
        make_goal(Pred, Args, Goal).
make_goal(Pred, Args, Goal) :-
        Goal =.. [ Pred | Args ].

%% Reinitialises the library
reinit :-
        remove_all_constraints(arity(_,_)),
        remove_all_constraints(arity(_,_,_)),
        repeat,
        (retract((find_mapping(_,_,_) :- !)) -> fail ; true),!.
