%% tclp__arrays.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__arrays.pl,v 1.1 2002/10/17 15:30:17 coquery Exp $

%% This modules provide fixed length arrays and matrix

:- module(tclp__arrays,[ create_array/2,
                         array_size/2,
                         array_element_at/3,
                         array_iteri_goal/5,
                         array_iter_goal/4,
                         array_iteri/2,
                         array_iter/2,
                         create_matrix/3,
                         create_matrix/2,
                         matrix_size/3,
                         matrix_element_at/4,
                         matrix_iterij_goal/6
                       ]).

:- use_module(library(lists)).

%% arrays are represented by terms of the form a(...) where ... are elements
%% of the array or of the form b(...) where ... are matrix or arrays
%% (because of the limitation of the number of arguments in a term)
%% arrays are encapsulated in a term (which permits to give information
%% such as the array's size)

%% creates an core array of size n
create_core_array(N,Array,Height) :-
        create_core_containers(N, Containers),
        build_layers(Containers,1,Array,Height).

%% creates a list of containers (a(...)) such that all theses containers form
%% the final containers of the array
create_core_containers(N,Containers) :-
        (
          N =< 255
        ->
          functor(Last,a,N),
          Containers = [ Last ]
        ;
          functor(Current,a,255),
          Containers = [ Current | Rest ],
          N1 is N - 255,
          create_core_containers(N1,Rest)
        ).

%% creates the tree structure by accumulating creating a list of super nodes
%% b(...) containing the nodes in the current list
build_one_layer([],[]) :- !.
build_one_layer(Nodes, [ SuperNode | SuperNodes ]) :-        
        partition_layer(255, Nodes, First, Rest),
        SuperNode =.. [ b | First ],
        build_one_layer(Rest,SuperNodes).

%% cuts the first argument in two lists, the first list having length N
partition_layer(0, List, [], List) :- !.
partition_layer(_, [], [], []) :- !.
partition_layer(N, [X|List], [X|First], Rest) :-
        N1 is N - 1,
        partition_layer(N1,List,First,Rest).

%% creates all the layers for a given list of containers, and returns
%% the height of the tree, a(...) corresponding to height = 1
build_layers([Node],Height,Node,Height) :- !.
build_layers(Nodes,Height,Node,FinalHeight) :-
        build_one_layer(Nodes, NewNodes),
        Height1 is Height + 1,
        build_layers(NewNodes,Height1,Node,FinalHeight).

%% creates an array of size N
create_array(N,array(N,H,CoreArray)) :-
        create_core_array(N,CoreArray,H).

%% get the size of an array
array_size(array(N,_,_), N).

%% get an element in an array
array_element_at(array(_,H,Core),Index,Element) :-
        (var(Core),write(Core),nl;true),
        core_element_at(Core,H,Index,Element).

%% get an element in a core array given its index and the height of the array
core_element_at(Core,1,Index,Element) :- !,
        (var(Core),write(Core),nl;true),
        arg(Index,Core,Element).
core_element_at(Core,H,Index,Element) :-
        (var(Core),write(Core),nl;true),
        Offset is integer(255 ** (H-1)),
        SubTreeIndex is Index // Offset + 1,
        SubIndex is Index mod Offset,
        arg(SubTreeIndex, Core, SubTree),
        H1 is H - 1,
        core_element_at(SubTree, H1, SubIndex, Element).

%% calls the predicate p with the given list as arguments
%% handles module stuff
predicate_call(Module:P,Args) :- !,
        Called =.. [ P | Args ],
        call(Module:Called).
predicate_call(P,Args) :-
        Called =.. [ P | Args ],
        call(Called).

%% array_iteri(Array,p) iters the given the goal p(i,e) for all i,e such that
%% i is the index of e in the array
array_iteri(array(_,_,Core),Predicate) :-
        core_iteri(Core,1,Predicate,_).

core_iteri(Node,FirstIndex,Predicate,LastIndex) :-
        Node =.. [ a | Args ],!,
        elements_iteri(Args, FirstIndex, Predicate, LastIndex).
core_iteri(Node,FirstIndex,Predicate,LastIndex) :-
        Node =.. [ b | Args ],!,
        nodes_iteri(Args, FirstIndex, Predicate, LastIndex).

elements_iteri([], Index, _, Index).
elements_iteri([ Element | Rest ], Index, Predicate, LastIndex) :-
        predicate_call(Predicate,[Index,Element]),
        Index1 is Index+1,
        elements_iteri(Rest, Index1, Predicate, LastIndex).

nodes_iteri([], Index, _, Index).
nodes_iteri([Node|Rest], FirstIndex, Predicate, LastIndex) :-
        core_iteri(Node, FirstIndex, Predicate, Index),
        nodes_iteri(Rest, Index, Predicate, LastIndex).
        
%% array_iter(Array,p) iters the given the goal p(e) for all e in the array
array_iter(array(_,_,Core),Predicate) :-
        core_iter(Core,Predicate).

core_iter(Node,Predicate) :-
        Node =.. [ a | Args ],!,
        elements_iter(Args, Predicate).
core_iter(Node,Predicate) :-
        Node =.. [ b | Args ],!,
        nodes_iter(Args, Predicate).

elements_iter([],_).
elements_iter([Element|Rest],Predicate) :-
        predicate_call(Predicate,[Element]),
        elements_iter(Rest,Predicate).

nodes_iter([],_).
nodes_iter([Node|Rest], Predicate) :-
        core_iter(Node, Predicate),
        nodes_iter(Rest, Predicate).

%% array_iteri_goal(Array,I,Element,Goal,NoDuplicate) for each I,Element such
%% that I is the index of Element in Array, calls a copy of Goal with Element
%% and Index instanciated to the correct values (anything in NoDuplicate is
%% no copied but unified in the copy)

array_iteri_goal(array(_,_,Core),I,E,Goal,NoDuplicate) :-
        core_iteri_goal(Core,1,I,E,Goal,NoDuplicate,_).

core_iteri_goal(Node,FirstIndex,I,E,Goal,NoDuplicate,LastIndex) :-
        Node =.. [ a | Args ],!,
        elements_iteri_goal(Args, FirstIndex, I,E,Goal,NoDuplicate, LastIndex).
core_iteri_goal(Node,FirstIndex,I,E,Goal,NoDuplicate,LastIndex) :-
        Node =.. [ b | Args ],!,
        nodes_iteri_goal(Args, FirstIndex, I,E,Goal,NoDuplicate, LastIndex).

elements_iteri_goal([], Index, _,_,_,_, Index).
elements_iteri_goal([ Element | Rest ],Index,I,E,Goal,NoDuplicate,LastIndex) :-
        copy_term([I,E,Goal,NoDuplicate],[Index,Element,Called,NoDuplicate]),
        call(Called),
        Index1 is Index+1,
        elements_iteri_goal(Rest, Index1, I,E,Goal,NoDuplicate, LastIndex).

nodes_iteri_goal([], Index, _,_,_,_, Index).
nodes_iteri_goal([Node|Rest], FirstIndex, I,E,Goal,NoDuplicate, LastIndex) :-
        core_iteri_goal(Node, FirstIndex, I,E,Goal,NoDuplicate, Index),
        nodes_iteri_goal(Rest, Index, I,E,Goal,NoDuplicate, LastIndex).
        
%% array_iter_goal/4 : same as array_iteri_goal/5 but without the index

array_iter_goal(array(_,_,Core),E,Goal,NoDuplicate) :-
        core_iter_goal(Core,E,Goal,NoDuplicate).

core_iter_goal(Node,E,Goal,NoDuplicate) :-
        Node =.. [ a | Args ],!,
        elements_iter_goal(Args,E,Goal,NoDuplicate).
core_iter_goal(Node,E,Goal,NoDuplicate) :-
        Node =.. [ b | Args ],!,
        nodes_iter_goal(Args,E,Goal,NoDuplicate).

elements_iter_goal([],_,_,_).
elements_iter_goal([ Element | Rest ],E,Goal,NoDuplicate) :-
        copy_term([E,Goal,NoDuplicate],[Element,Called,NoDuplicate]),
        call(Called),
        elements_iter_goal(Rest,E,Goal,NoDuplicate).

nodes_iter_goal([],_,_,_).
nodes_iter_goal([Node|Rest],E,Goal,NoDuplicate) :-
        core_iter_goal(Node,E,Goal,NoDuplicate),
        nodes_iter_goal(Rest,E,Goal,NoDuplicate).

%% creates a matrix of size M x N
create_matrix(M,N,Matrix) :-
        create_array(M,Matrix),
        array_iter_goal(Matrix, Column, create_array(N,Column), N).

%% creates a square matrix of size N
create_matrix(N,Matrix) :-
        create_matrix(N,N,Matrix).

%% matrix size
matrix_size(Matrix,M,N) :-
        array_size(Matrix,M),
        array_element_at(Matrix,1,Column),
        array_size(Column,N).

%% get an element in a matrix
matrix_element_at(Matrix,I,J,Element) :-
        array_element_at(Matrix,I,Column),
        array_element_at(Column,J,Element).

%% iters a goal on a matrix
matrix_iterij_goal(Matrix, I, J, Element, Goal, NoDuplicate) :-
        array_iteri_goal(Matrix, I, Column,
                         array_iteri_goal(Column,J,Element,Goal,NoDuplicate),
                         NoDuplicate).
